import { CalculatorType, ComputeNodeType, Nub, ParamDefinition, Props, Observer } from "jalhyd";

import { FormulaireElement } from "../formulaire-element";
import { NgParameter, ParamRadioConfig } from "../ngparam";
import { Field } from "../field";
import { StringMap } from "../../stringmap";
import { FormulaireNode } from "../formulaire-node";
import { ServiceFactory } from "../../services/service-factory";
import { ParamService } from "../../services/param/param.service";
import { FieldSet } from "../fieldset";
import { FieldsetContainer } from "../fieldset-container";
import { SelectField } from "../select-field";
import { DeepFieldsetIterator } from "../form-iterator/deep-fieldset-iterator";
import { DeepFormulaireElementIterator } from "../form-iterator/deep-element-iterator";
import { TopFormulaireElementIterator } from "../form-iterator/top-element-iterator";
import { CalculatorResults } from "../../results/calculator-results";

/**
 * classe de base pour tous les formulaires
 */
export abstract class FormulaireDefinition extends FormulaireNode implements Observer {
    /**
     * nom de la calculette
     */
    private _calculatorName: string;

    /**
     * Nub courant
     */
    protected _currentNub: Nub;

    /**
     * propriétés par défaut (lues si _currentNub === undefined )
     */
    private _props = {};

    /**
     * fichier de configuration
     */
    private _jsonConfig: {};

    /**
     * aide en ligne
     */
    private _helpLink: string;

    protected _paramService: ParamService;

    constructor() {
        super(undefined);
        this._paramService = ServiceFactory.instance.paramService;
    }

    public get calculatorType(): CalculatorType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["calcType"];
    }

    public get nodeType(): ComputeNodeType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["nodeType"];
    }

    public get calculatorName() {
        return this._calculatorName;
    }

    public set calculatorName(name: string) {
        this._calculatorName = name;
        this.notifyObservers({
            "action": "nameChanged",
            "name": name
        },
            this);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    public get defaultProperties() {
        return this._props;
    }

    public initNub(props?: {}) {
        this._currentNub = this.createNub(props === undefined ? this.defaultProperties : props);
    }

    public get currentNub(): Nub {
        return this._currentNub;
    }

    public set currentNub(n: Nub) {
        const nubCalcType = (n.properties as Props).getPropValue("calcType");
        if (this._props["calcType"] !== nubCalcType) {
            throw new Error(
                `Nub ${n.properties["calcType"]} incompatible avec le formulaire ${this._calculatorName} (${this._props["calcType"]})`
            );
        }
        this._currentNub = n;
    }

    protected createNub(params: Props | {}): Nub {
        const props = params instanceof Props ? params : new Props(params);
        return this._paramService.createNub(props);
    }

    protected replaceCurrentNub(params: Props) {
        this.currentNub = this._paramService.replaceNub(this._currentNub, params);
    }

    protected replaceNub(sn: Nub, params: Props): Nub {
        return this._paramService.replaceNub(sn, params);
    }

    protected deleteNub(sn: Nub) {
        this._paramService.deleteNub(sn);
    }

    protected initParse() {
    }

    protected parseOptions(json: {}) {
        const dnt = json["defaultNodeType"];
        this._props["nodeType"] = dnt === undefined ? ComputeNodeType.None : ComputeNodeType[dnt];
        this.helpLink = json["help"];
    }

    protected completeParse(json: {}) {
    }

    public getOption(json: {}, option: string): string {
        if (json["type"] === "options") {
            return json[option];
        }

        return undefined;
    }

    public afterParseFieldset(fs: FieldSet) {
    }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}): FieldSet {
        const res: FieldSet = new FieldSet(parent);
        res.setNub(this._currentNub, false);
        this.kids.push(res);
        return res;
    }

    public moveFieldsetUp(fs: FieldSet) {
    }

    public moveFieldsetDown(fs: FieldSet) {
    }

    public removeFieldset(fs: FieldSet) {
    }

    private parse_fieldset(json: {}) {
        const fs = this.createFieldset(this, json);
        fs.parseConfig(json);
        this.afterParseFieldset(fs);
    }

    private parse_template_container(json: {}, templates: any[]) {
        const fsc: FieldsetContainer = new FieldsetContainer(this);
        fsc.parseConfig(json, templates);
        this.kids.push(fsc);
    }

    public parseDependencies(json: {}) {
        // tslint:disable-next-line:forin
        for (const conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                // field set
                case "fieldset":
                case "template_container":
                    for (const k of this.kids) {
                        if (k.id === conf["id"]) {
                            k.parseDependencies(conf);
                            break;
                        }
                    }
                    break;

                case "fieldset_template":
                    for (const k of this.kids) {
                        if (k instanceof FieldsetContainer) {
                            k.parseDependencies(conf);
                        }
                    }
                    break;
            }
        }
    }

    /**
     * 1ère passe d'analyse de la configuration
     */
    public preparseConfig(json: {}) {
        this._jsonConfig = json;

        this.initParse();

        // analyse des options globales
        // il est utile de le faire avant le reste pour les calculettes utilisant
        // des sections (id des selects type de section/variable à calculer)

        // tslint:disable-next-line:forin
        for (const conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            if (type === "options") {
                this.parseOptions(conf);
                break;
            }
        }
    }

    /**
     * 2ème passe d'analyse de la configuration
     */
    public parseConfig(json: {}) {
        if (json !== undefined) {
            this._jsonConfig = json;
        }

        // analyse des éléments du formulaire
        const templates: any[] = [];

        // tslint:disable-next-line:forin
        for (const conf_index in this._jsonConfig) {
            const conf = this._jsonConfig[conf_index];
            const type: string = conf["type"];

            switch (type) {
                case "fieldset":
                    this.parse_fieldset(conf);
                    break;

                case "fieldset_template":
                    templates.push(conf);
                    break;

                // options globales
                case "options":
                    break;

                case "template_container":
                    this.parse_template_container(conf, templates);
                    break;

                default:
                    throw new Error(`type d'objet de calculette ${type} non pris en charge`);
            }
        }

        this.completeParse(this._jsonConfig);

        // console.log("-----");

        // for (const n of Session.getInstance().NubIterator) {
        //     console.log(n.nub);
        //     for (const p of n.nub.parameterIterator)
        //         console.log(`${p.symbol} uid  ${p.uid} props ${n.properties} mode ${p.valueMode} val ${p.uncheckedValue}`);
        // }

        // logObject(this._fieldSets, "fieldsets");
        // logObject(this._dependencies, "dependences");
        this.parseDependencies(this._jsonConfig);
    }

    public hasParameter(symbol: string): boolean {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return true;
                }
            }
        }
        return false;
    }

    public getParamFromSymbol(symbol: string): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return p;
                }
            }
        }
        return undefined;
    }

    public getDisplayedParamFromState(st: ParamRadioConfig): NgParameter {
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter) {
                if (p.radioState === st) {
                    return p;
                }
            }
        }
        return undefined;
    }

    public getDisplayedParamListFromState(st: ParamRadioConfig): NgParameter[] {
        const res = [];
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter && p.radioState === st) {
                res.push(p);
            }
        }
        return res;
    }

    public getFieldById(id: string): Field {
        const res = this.getFormulaireNodeById(id);
        if (res instanceof Field) {
            return res;
        }
        return undefined;
    }

    public getParameterValue(symbol: string): number {
        for (const fs of this.allFieldsets) {
            const p = fs.getNodeParameter(symbol);
            if (p !== undefined) {
                switch (p.radioState) {
                    case ParamRadioConfig.FIX:
                        return p.getValue();

                    case ParamRadioConfig.VAR:
                    case ParamRadioConfig.CAL:
                        return undefined;
                }
            }
        }

        throw new Error(`Formulaire.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
    }

    private notifyReset() {
        this.notifyObservers({
            "action": "resetForm"
        }, this);
    }

    /**
     * réinitialisation du formulaire suite à un changement d'une valeur, d'une option, ... :
     * effacement des résultats, application des dépendances, ...
     */
    public reset() {
        this.resetResults();
        this.applyDependencies();

        // prévenir les composants qu'il faut détecter les changements
        this.notifyReset();
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): any {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().value;
    }

    /**
     * retourne le label actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns label courant du select
     */
    public getSelectedLabel(selectFieldId: string): string {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().label;
    }

    public applyDependencies() {
        for (const fe of this.topFormElements) {
            fe.applyDependencies();
        }
    }

    public abstract resetResults();
    public abstract doCompute();
    public abstract get hasResults(): boolean;
    public abstract get results(): CalculatorResults[];

    public updateLocalisation(localisation: StringMap) {
        for (const fe of this.topFormElements) {
            fe.updateLocalisation(localisation);
        }

        if (this.hasResults) {
            this.doCompute(); // pour mettre à jour la langue
        }
    }

    public get helpLink(): string {
        return this._helpLink;
    }

    public set helpLink(h: string) {
        this._helpLink = h;
    }

    public isDisplayed(id: string) {
        return (<FormulaireElement>this.getFormulaireNodeById(id)).isDisplayed;
    }

    public get isValid(): boolean {
        let res = true;
        for (const fs of this.allFieldsets) {
            if (fs.isDisplayed) {
                res = res && fs.isValid;
            }
        }
        return res;
    }

    /**
     * gestion d'un clic sur les radios
     */
    public onRadioClick(info: any) {
        this.reset();
    }

    /**
     * itère sur tous les FieldSet
     */
    public get allFieldsets(): IterableIterator<FieldSet> {
        return new DeepFieldsetIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement
     */
    public get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement de 1er niveau
    */
    private get topFormElements(): IterableIterator<FormulaireElement> {
        return new TopFormulaireElementIterator(this);
    }

    /**
     * sérialisation en JSON
     */
    public JSONserialise(): {} {
        const res = {};
        res["id"] = this.calculatorName;
        res["uid"] = this.uid;
        res["props"] = (this._currentNub.properties as Props).props;
        res["elements"] = this.serialiseKids();
        return { "form": res };
    }

    private deserialiseFieldset(elements: {}): FieldSet {
        const res: FieldSet = this.getFormulaireNodeById(elements["id"]) as FieldSet;
        res.deserialiseJSON(elements);
        return res;
    }

    private deserialiseFieldsetContainer(elements: {}): FieldsetContainer {
        const res: FieldsetContainer = this.getFormulaireNodeById(elements["id"]) as FieldsetContainer;
        res.deserialiseJSON(elements);
        return res;
    }

    private deserialiseElement(element: {}) {
        const keys = Object.keys(element);
        if (keys.length !== 1) {
            throw new Error(`session file : invalid form object '${element}'`);
        }

        switch (keys[0]) {
            case "fieldset":
                this.deserialiseFieldset(element[keys[0]]);
                break;

            case "fieldset_container":
                this.deserialiseFieldsetContainer(element[keys[0]]);
                break;

            default:
                throw new Error(`session file : invalid key '${keys[0]}' in form object`);
        }
    }

    /**
     * désérialisation depuis JSON
     */
    public deserialiseJSON(elements: {}) {
        // tslint:disable-next-line:forin
        for (const k in elements) {
            switch (k) {
                case "id":
                    this._calculatorName = elements[k];
                    break;

                case "uid":
                case "props":
                    break;

                case "elements":
                    for (const e of elements[k]) {
                        this.deserialiseElement(e);
                    }
                    break;

                default:
                    throw new Error(`session file : invalid key '${k}' in form object`);
            }
        }
    }

    //  interface Observer

    public update(sender: any, data: any) {
    }
}
