import { cLog, Result, ResultElement, ParamValues, CourbeRemousParams, ParamDefinition, ParamDomainValue } from "jalhyd";

import { CalculatorResults } from "./calculator-results";
import { VarResults } from "./var-results";
import { NgParameter } from "../formulaire/ngparam";

export class RemousResults extends CalculatorResults {
    private _result: Result;

    private _xValues: ParamDefinition;

    /**
     * hauteur de berge
     */
    private _hautBerge: number;

    /**
     * pente du fond
     */
    private _penteFond: number;

    /**
     * hauteur normale
     */
    private _hautNormale: ResultElement;

    /**
     * hauteur critique
     */
    private _hautCritique: ResultElement;

    /**
     * présence de données pour la courbe fluviale
     */
    private _hasFlu: boolean;

    /**
     * présence de données pour la courbe torrentielle
     */
    private _hasTor: boolean;

    /**
     * présence de données pour le paramètre supplémentaire
     */
    private _hasExtra: boolean;

    /**
     * résultats variables pour l'affichage d'une table de résultats numériques
     */
    private _varResults: VarResults;

    /**
     * le paramètre supplémentaire est affiché dans un graphe séparé
     */
    private _extraGraph: boolean;

    /**
     * titre de la colonne du paramètre supplémentaire
     */
    private _extraParamSymbol: string;

    /**
     * journal de calcul
     */
    private _log: cLog;

    constructor() {
        super();
        this.reset();
    }

    public reset() {
        this._result = undefined;
        this._log = new cLog();
        this._hautBerge = undefined;
        this._penteFond = undefined;
        this._hautNormale = undefined;
        this._hautCritique = undefined;
        this._extraParamSymbol = undefined;
        this._hasFlu = false;
        this._hasTor = false;
        this._hasExtra = false;
        this._extraGraph = false;
    }

    public set parameters(p: CourbeRemousParams) {
        // pente du fond
        this._penteFond = p.map.If.v;

        // hauteur de berge
        this._hautBerge = p.map.YB.v;

        // longueur du bief
        const Long = p.map.Long.v;

        // pas d'espace
        const Dx = p.map.Dx.v;

        // série de valeurs de X
        this._xValues = new ParamDefinition("Abscisse", ParamDomainValue.POS_NULL);
        this._xValues.paramValues.setValues(0, Long, Dx);
    }

    public get log() {
        return this._log;
    }

    public get varResults(): VarResults {
        return this._varResults;
    }

    public get result(): Result {
        return this._result;
    }

    public set result(r: Result) {
        this._result = r;
    }

    public update() {
        this._hasFlu = false;
        this._hasTor = false;
        this._hasExtra = false;

        for (const re of this._result.resultElements) {
            if (!this._hasFlu && re.getExtraResult("flu")) {
                this._hasFlu = true;
            }
            if (!this._hasTor && re.getExtraResult("tor")) {
                this._hasTor = true;
            }
            if (!this._hasExtra && re.getExtraResult(this.extraParamSymbol)) {
                this._hasExtra = true;
            }
        }

        this._log.addLog(this._result.globalLog);

        this._varResults = new VarResults();
        this._varResults.variatedParameter = new NgParameter(this._xValues, undefined);
        this._varResults.calculatedParameter = new NgParameter(new ParamDefinition("Ligne d'eau", ParamDomainValue.POS_NULL), undefined);
        this._varResults.result = this._result;
        const keys = [];
        if (this._hasFlu) {
            keys.push("flu");
        }
        if (this._hasTor) {
            keys.push("tor");
        }
        if (this._hasExtra) {
            keys.push(this.extraParamSymbol);
        }
        this._varResults.extraResultKeys = keys;
        this._varResults.update(true);
    }

    public get extraParamSymbol(): string {
        return this._extraParamSymbol;
    }

    public set extraParamSymbol(l: string) {
        this._extraParamSymbol = l;
    }

    public get hautBerge() {
        return this._hautBerge;
    }

    public get penteFond() {
        return this._penteFond;
    }

    public get hautNormale() {
        return this._hautNormale;
    }

    public set hauteurNormale(v: ResultElement) {
        this._hautNormale = v;
    }

    public get hautCritique() {
        return this._hautCritique;
    }

    public set hauteurCritique(v: ResultElement) {
        this._hautCritique = v;
    }

    public get hasExtra() {
        return this._hasExtra;
    }

    public get extraGraph() {
        return this._extraGraph;
    }

    public set extraGraph(b: boolean) {
        this._extraGraph = b;
    }

    public get hasResults(): boolean {
        return this._hautBerge !== undefined ||
            this._penteFond !== undefined ||
            this._hautNormale !== undefined ||
            this._hautCritique !== undefined ||
            (this._result && this._result.ok);
    }

    /**
     * @return true si il existe au moins une données dans la série fluviale
     */
    public get hasFluData(): boolean {
        return this._hasFlu;
    }

    /**
     * @return true si il existe au moins une données dans la série torrentielle
     */
    public get hasTorData(): boolean {
        return this._hasTor;
    }
}
